import pytest
from aiohttp import ClientSession
from pyaffalddk import GarbageCollection


@pytest.mark.asyncio
async def test_Aalborg(capsys):
    with capsys.disabled():
        session = ClientSession()
        gc = GarbageCollection('Aalborg', session=session)

        address = await gc.get_address_id('9000', 'Boulevarden', '13')
        add = {'address_id': '139322', 'kommunenavn': 'Aalborg', 'vejnavn': 'Boulevarden', 'husnr': '13'}
        # print(address.__dict__)
        assert address.__dict__ == add

        pickups = await gc.get_pickup_data(address.address_id)
        assert len(pickups) > 1
        assert 'next_pickup' in pickups


@pytest.mark.asyncio
async def test_Odense(capsys):
    with capsys.disabled():
        session = ClientSession()
        gc = GarbageCollection('Odense', session=session)

        address = await gc.get_address_id('5000', 'Flakhaven', '2')
        # print(address.__dict__)
        add = {'address_id': '112970', 'kommunenavn': 'Odense', 'vejnavn': 'Flakhaven', 'husnr': '2'}
        assert address.__dict__ == add

        pickups = await gc.get_pickup_data(address.address_id)
        assert len(pickups) > 1


@pytest.mark.asyncio
async def test_Aarhus(capsys):
    with capsys.disabled():
        session = ClientSession()
        gc = GarbageCollection('Aarhus', session=session)

        address = await gc.get_address_id('8000', 'Rådhuspladsen', '2')
        # print(address.__dict__)
        add = {'address_id': '07517005___2_______', 'kommunenavn': 'Aarhus', 'vejnavn': 'Rådhuspladsen', 'husnr': '2'}
        assert address.__dict__ == add

        pickups = await gc.get_pickup_data(address.address_id)
        assert len(pickups) > 1
        assert 'next_pickup' in pickups


@pytest.mark.asyncio
async def test_Kbh(capsys):
    with capsys.disabled():
        session = ClientSession()
        gc = GarbageCollection('København', session=session)

        address = await gc.get_address_id('1550', 'Rådhuspladsen', '1')
        # print(address.__dict__)
        add = {'address_id': 'a4e9a503-c27f-ef11-9169-005056823710', 'kommunenavn': 'København', 'vejnavn': 'Rådhuspladsen', 'husnr': '1'}
        assert address.__dict__ == add

        pickups = await gc.get_pickup_data(address.address_id)
        assert len(pickups) > 1
        assert 'next_pickup' in pickups



"""Constants for the pyaffalddk integration."""
from __future__ import annotations

NON_SUPPORTED_ITEMS = [
    "Asbest",
    "Beholderservice",
    "Bestillerordning",
    "Henteordning for grene",
    "Trærødder og stammer",
    "Ingen tømningsdato fundet!",
    "Skal tilmeldes",
    "Bestil afhentning",
    "Farligt affald (skal bestilles)",
    "Storskrald (skal bestilles)",
    "Haveaffald (skal bestilles)",
    "Beholderudbringning",
]

SUPPORTED_ITEMS = {
    "restaffaldmadaffald": [
        "Restaffald-Madaffald",
        "Rest/mad",
        "Restaffald",
        "Rest - Mad",
        "Rest-/Madaffald",
        "Mad- og restaffald",
        "Mad/rest",
        "Rest-/madaffald",
        "Mad/Rest",
        "Mad og restaffald",
        "Mad- og restaffald",
        "Mad-/ og restaffald",
        "Rest / Mad",
        "Rest Mad",
        "Energispand - Obligatorisk min. 1 spand",
        "Rest/madaffald",
        "Energibeholder (mad/rest)",
        "Rest- og Madaffald",
    ],
    "madaffald": [
        "Madaffald",
        "Bioposer",
    ],
    "batterier": ["Batterier"],
    "dagrenovation": ["Dagrenovation"],
    "elektronik": ["Elektronik"],
    "glas": [
        "Industri Genbrugeligt",
        "Glas",
    ],
    "metalglas": [
        "Glas og metal",
        "Metal-Glas",
        "Glas/Metal",
        "Metal og Glas",
        "Glas / Metal",],
    "pappi": [
        "Plast MDK og papir",
        "PAPPI",
        "Plast/MD-karton/PP",
        "Plast/Papir",
        "Papir/Plast og kartoner",
        "Plast og Mad- & drikkekartoner/Papir",
        "Plast, Mad- og drikkekartoner, Småt Pap og Papir",
        "Papir & Pap / Plast & MDK",
    ],
    "farligtaffald": [
        "Farligt affald",
        "Miljøkasser",
        "Farligt Affald, Rød boks",
        "Genbrugsbilen",
    ],
    "farligtaffaldmiljoboks": [
        "Farligt affald/Miljøboks",
        "Papir og glas/dåser",
        "Miljøboks",
    ],
    "flis": [
        "Flis",
        "Flishugning",
    ],
    "genbrug": [
        "Genbrug",
        "Genbrugsøer",
        "Genanvendeligt",
        "Genbrug 240 L",
        "Genbrugshal (1 stk.)",
    ],
    "jern": [
        "Jern",
        "Metal",
    ],
    "juletrae": [
        "Juletræ",
    ],
    "papir": [
        "Papir",
        "Papir 660 L",
        "Papir // Tekstilaffald",
    ],
    "papirglasdaaser": [
        "Papir og glas/dåser",
    ],
    "papirmetal": [
        "Papir & Metal",
        "Papir/metal",
        "Metal/papir",
        "Papir og Metal",
        "Pap & metal",
    ],
    "pap": [
        "Pap",
    ],
    "plastmetal": [
        "Plast & Metal",
        "Plast, metal & MDK 660L",
        "Plast/ Metal",
        "Plast/MDK/Metal",
        "Plast-metal",
        "Plast/Metal",
        "Plast, Metal, MDK",
    ],
    "restaffald": [
        "Restaffald // Batterier og Småt Elektronik",
        "Industri Restaffald",
    ],
    "storskrald": [
        "Storskrald",
        "Stort elektronik",
        "Storskrald og genbrug",
        "Pap og Storskrald",
        "Stort indbo og Indendørs træ",
    ],
    "storskraldogtekstilaffald": [
        "Storskrald og tekstilaffald",
    ],
    "haveaffald": [
        "Haveaffald, flishugning",
        "Haveaffald",
        "Trærødder og stammer",
        "Haveaffald - Frivilligt",
        "Haveaffald 140 L",
        "Frivillig Haveaffald tømmes fra mandag til onsdag",
    ],
    "papirglas": [
        "Papir og pap/Glas",
        "Glas-papir",
        "Papir, Pap/Glas",
        "Papir & glas",
    ],
    "plastmadkarton": [
        "Plast/MDK",
        "Plast & mad- og drikkekartoner",
        "Plast/ Mad- og drikkekartoner",
        "Plast og MDK",
    ],
    "pappapirglasmetal": [
        "Pap, papir, tekstil & metal og glas",
        "4delt beholder",
        "Ressourcebeholder (pap/papir og glas/metal)",
        "Papir/pap og glas/metal",
    ],
    "plastmetalmadmdk": [
        "Plast/Metal/Mad- & drikkekartoner",
        "Plast, mad- og drikkekartoner og metal",
    ],
    "plastmdkglasmetal": [
        "Plast og drikkekarton – Glas og metal",
    ],
    "pappapir": [
        "Pap/Papir",
        "Papir/Pap",
        "Papir og Pap",
        "Papir, pap og tekstilaffald",
    ],
    "tekstil": [
        "Tekstiler",
        "Standplads",
        "Tekstilaffald",
        "Miljøkasse/tekstiler",
        "Tekstil",
        "Tekstilposer (1 stk.)",
    ],
    "glasplast": [
        "Glas/Plast og MDK",
    ],
    "plastmetalpapir": [
        "Plast/Metal/Papir ",
        "Genbrugsspand - Obligatorisk min. 1 spand",
    ],
    "plast": ["Plast - Obligatorisk min. 1 spand", "Plast"],
    "restplast": [
        "Rest + plast/MDK",
    ],
}

NON_MATERIAL_LIST = [
    "Pap bundtet, havebolig (1 stk.)",
]

MATERIAL_LIST = {
    "restaffaldmadaffald": [
        "Mad- og restaffald",
        "*190 l rød, 14 dags tømning (1 stk.)",
    ],
    "batterier": [
        "Batterier i pose på låg (1 stk.)",
    ],
    "dagrenovation": [
        "240 l. container 1 x ugentligt (12 stk.)",
        "*140 l grøn (1 stk.)",
    ],
    "elektronik": [
        "Småt elektronik i pose på låg (1 stk.)",
    ],
    "glas": [
        "240 L glas (1 stk.)",
        "Glas, 660 liter (2 stk.)",
        "400 liter - Glas - Tømmes hver 8. uge (1 stk.)",
        "Glas",
        "Cube 1500 L (Flasker) (1 stk.)",
    ],
    "metalglas": [
        "Glas/metal (1 stk.)",
        "240 l Glas/metal Egenløsning (1 stk.)",
        "Metal/Glas 240 l - 2-kammer (1 stk.)",
        "Glas/metal, afstand over 5 meter (1 stk.)",
        "240 L - Metal og Glas (1 stk.)",
        "240 l glas/metal - 4 tømninger (1 stk.)",
        "240 liter - Metal/Glas - tømmes hver 8. uge (1 stk.)",
        "Metal/glas 240 l (hver 12. uge) (1 stk.)",
        "240 L. Glas/Metal - 4 tømninger (1 stk.)",
        "240 l Glas/metal Egenløsning (Privat) (1 stk.)",
        "Glas & Metal",
    ],
    "pappi": [
        "Plast/papir (1 stk.)",
        "370 l Plast+MDK/Papir - Egenløsning (1 stk.)",
        "240 l Plast+MDK/Papir - Egenløsning (1 stk.)",
        "Papir/Plast og MDK 240 l - 2-kammer (1 stk.)",
        "Plast/papir, afstand over 5 meter (1 stk.)",
        "240 L - Plast og Papir (1 stk.)",
        "240L genbrug (1 stk.)",
        "240 l PMDK/PP 3 ugers tømning (1 stk.)",
        "240 liter - Plast, MDK, papir, pap – tømmes hver 2. uge (1 stk.)",
        "PP/MDK-plast 240 l (hver 2. uge) (1 stk.)",
        "240 L. PMDK/PP 3 ugers tømning (1 stk.)",
        "370 l Plast+MDK/Papir - Egenløsning (Privat) (1 stk.)",
        "240 l Plast+MDK/Papir - Egenløsning (Privat) (1 stk.)",
        "Papir & småt pap",
    ],
    "farligtaffald": [
        "",
    ],
    "farligtaffaldmiljoboks": [
        "Miljøboks",
        "Miljøkasse (1 stk.)",
        "Miljøskab (1 stk.)",
        "Rød kasse henteordning (1 stk.)",
        "*Miljøkasse (1 stk.)",
    ],
    "flis": [""],
    "jern": [
        "240 L metal (1 stk.)",
        "Metal, 660 liter (1 stk.)",
        "240 liter - Metal - Tømmes hver 8. uge (1 stk.)",
        "Metal",
    ],
    "genbrug": [
        "Tekstiler",
        "Genbrug",
        "Genbrugsøer",
        "Genanvendeligt",
        "Genbrug 240 L",
        "Genbrugshal (1 stk.)",
    ],
    "papir": [
        "660 L papir (1 stk.)",
        "Papir, 660 liter (3 stk.)",
        "Papir",
    ],
    "papirmetal": [
        "Metal og papir 240 L (ejer.kommune) (1 stk.)",
    ],
    "pap": [
        "Pap 240 L (villa) (1 stk.)",
        "240 l Pap Egenløsning (1 stk.)",
        "Pap 240 L (1 stk.)",
        "Pap 240 L (ejer:kommune) (h) (1 stk.)",
        "Pap - 240 l.  (1 stk.)",
        "Indsamling af pap (løst og beholder) (1 stk.)",
        "240 L - Pap - Skel (1 stk.)",
        "Beholder til pap (1 stk.)",
        "770 L pap (1 stk.)",
        "Pap, 660 liter (3 stk.)",
        "400 liter - Pap og papir - Tømmes hver 2. uge (1 stk.)",
        "Pap",
        "240 l Pap Egenløsning (Privat) (1 stk.)",
    ],
    "plastmetal": [
        "Plast, småt metal & MDK - 240 l. (1 stk.)",
    ],
    "plastmdkglasmetal": [
        "240 L todelt genbrugsspand hver 4. uge (1 stk.)",
        "240L plast, MDK/metal, glas gl. (1 stk.)",
        "Genbrugsbeholder, 190 ltr (1 stk.)",
        "Genbrugsspand, 240L (2-delt) (1 stk.)",
    ],
    "plast": [
        "Plast og MDK 240 L (ejer.kommune) (1 stk.)",
        "Plast, 660 liter (2 stk.)",
        "Plastaffald henteordning (1 stk.)",
        "Plast",
    ],
    "storskrald": [
        "Storskrald - fortovsindsamling (1 stk.)",
        "Storskrald (1 stk.)",
        "Storskrald enfamiliehus (1 stk.)",
        "Storskrald",
        "Stort elektronik",
        "Storskrald og genbrug",
        "Pap og Storskrald",
        "Storskrald distrikt 8 (1 stk.)",
        "Møbler/indbo  (1 stk.)",
        "Elektronik, stort metal, keramik og porcelæn (1 stk.)",
        "Småt brændbart (1 stk.)",
    ],
    "storskraldogtekstilaffald": [""],
    "haveaffald": [
        "Gen-Skel 0-2 meter (1 stk.)",
        "Haveaffald henteordning (1 stk.)",
    ],
    "papirglas": [
        "Papir/glas - 240 l. fortovsindsamling (1 stk.)",
    ],
    "plastmadkarton": [
        "Genbrugsbeholder PMDK/MG-240L/3uge (1 stk.)",
        "Genbrug henteordning (1 stk.)",
        "660 L plast, mad- og drikkekartoner (1 stk.)",
        "240 l genbrug - blåt låg (1 stk.)",
        "660 liter - Plast/MDK - Tømmes hver 2. uge (1 stk.)",
        "Plast og mad- & drikkekartoner",
        "Mad- og drikkekartoner",
    ],
    "papirglasdaaser": [
        "240 L 2-delt Papir/glas-dåser en-familie (1 stk.)",
        "*240 L 2-delt Papir/glas-dåser en-familie (1 stk.)",
    ],
    "pappapirglasmetal": [
        "240L metal, glas, plast/papir gl. (1 stk.)",
        "240 l genbrug 2-kammer (2023) (1 stk.)",
        "4-kammer (240 l) (1 stk.)",
        "240 L 2-delt Papir/glas-dåser en-familie (1 stk.)",
        "240 l delt genbrug - grønt låg (1 stk.)",
        "Glas & metal og papir & småt pap",
    ],
    "plastmetalmadmdk": [""],
    "pappapir": [
        "Genbrugsbeholder PP-240L/6uge (1 stk.)",
        "240L papir, pap (1 stk.)",
    ],
    "tekstil": [
        "Tekstil pose (1 stk.)",
        "Tekstilpose tømning (1 stk.)",
        "Tekstilaffald (1 stk.)",
        "Tekstilpose 25 L (1 stk.)",
        "Tekstilaffald (i lukket pose) (1 stk.)",
    ],
    "glasplast": [""],
    "plastmetalpapir": [
        "*140 l grøn, 14 dags tømning (1 stk.)",
    ],
    "papirglasmetalplast": [
        "4-kammer (370 l) (1 stk.)",
        "Pap og papir/metal, glas og hård plast - 240 L (1 stk.)",
        "240 l genbrug låg i låg (1 stk.)",
        "Pap og papir/metal, glas og hård plast - 240 L (EX) (1 stk.)",
    ],
    "madaffald": [
        "140 L madaffald (1 stk.)",
        "*190 l rød (1 stk.)",
        "Madaffald",
    ],
    "restaffald": [
        "9505: 770 L restaffald uge (2 stk.)",
        "Restaffald",
    ],
    "juletrae": [
        "Juletræer (1 stk.)",
        "Juletræ",
    ],
}

ICON_LIST = {
    "batterier": "mdi:battery",
    "elektronik": "mdi:power-plug",
    "restaffaldmadaffald": "mdi:trash-can",
    "dagrenovation": "mdi:trash-can",
    "glas": "mdi:bottle-soda",
    "metalglas": "mdi:glass-fragile",
    "pappi": "mdi:trash-can",
    "farligtaffald": "mdi:recycle",
    "farligtaffaldmiljoboks": "mdi:recycle",
    "flis": "mdi:tree",
    "genbrug": "mdi:recycle",
    "jern": "mdi:bucket",
    "papir": "mdi:file",
    "papirmetal": "mdi:delete-empty",
    "pap": "mdi:note",
    "plastmetal": "mdi:trash-can-outline",
    "storskrald": "mdi:table-furniture",
    "storskraldogtekstilaffald": "mdi:table-furniture",
    "haveaffald": "mdi:leaf",
    "papirglas": "mdi:greenhouse",
    "papirglasmetalplast": "mdi:trash-can",
    "plastmadkarton": "mdi:trash-can",
    "papirglasdaaser": "mdi:trash-can",
    "pappapirglasmetal": "mdi:trash-can",
    "plastmdkglasmetal": "mdi:trash-can",
    "plastmetalmadmdk": "mdi:trash-can",
    "pappapir": "mdi:file",
    "tekstil": "mdi:recycle",
    "glasplast": "mdi:trash-can",
    "plastmetalpapir": "mdi:trash-can",
    "plast": "mdi:trash-can",
    "madaffald": "mdi:trash-can",
    "restaffald": "mdi:trash-can",
    "restplast": "mdi:trash-can",
    "juletrae": "mdi:pine-tree",
    "metal": "mdi:anvil",
}

NAME_LIST = {
    "batterier": "Batterier",
    "bioposer": "Bioposer",
    "elektronik": "Elektronik",
    "dagrenovation": "Dagrenovation",
    "farligtaffald": "Farligt affald",
    "farligtaffaldmiljoboks": "Farligt affald & Miljøboks",
    "flis": "Flis",
    "genbrug": "Genbrug",
    "glas": "Glas",
    "glasplast": "Glas, Plast & Madkartoner",
    "haveaffald": "Haveaffald",
    "jern": "Metal",
    "juletrae": "Juletræer",
    "madaffald": "Madaffald",
    "metal": "Metal",
    "metalglas": "Metal & Glas",
    "pap": "Pap",
    "papir": "Papir",
    "papirglas": "Papir, Pap & Glas",
    "papirglasdaaser": "Papir, Glas & Dåser",
    "papirglasmetalplast": "Papir, Glas, Metal & Plast",
    "papirmetal": "Papir & Metal",
    "pappapir": "Pap & Papir",
    "pappapirglasmetal": "Pap, Papir, Glas & Metal",
    "pappi": "Papir & Plast",
    "plast": "Plast",
    "plastmadkarton": "Plast & Madkarton",
    "plastmdkglasmetal": "Plast, Madkarton, Glas & Metal",
    "plastmetal": "Plast & Metal",
    "plastmetalmadmdk": "Plast, Metal, Mad & Drikkekartoner",
    "plastmetalpapir": "Plast, Metal & Papir",
    "restaffald": "Restaffald",
    "restaffaldmadaffald": "Rest & Madaffald",
    "restplast": "Restaffald & Plast/Madkartoner",
    "storskrald": "Storskrald",
    "storskraldogtekstilaffald": "Storskrald & Tekstilaffald",
    "tekstil": "Tekstilaffald",
}

NAME_ARRAY = list(NAME_LIST.keys())

MUNICIPALITIES_LIST = {
    "Aabenraa": ["aabenraa", "1"],
    "Aalborg": ["aalborg", "1"],
    "Aarhus": ["aarhus", "3"],
    "Albertslund": ["albertslund", "1"],
    "Allerød": ["allerod", "1"],
    "Billund": ["billund", "1"],
    "Bornholm": ["bofa", "1"],
    "Brøndby": ["brondby", "1"],
    "Brønderslev": ["bronderslev", "1"],
    "Dragør": ["dragoer", "1"],
    "Egedal": ["egedal", "1"],
    "Esbjerg": ["esbjerg", "1"],
    "Faxe": ["faxe", "1"],
    "Fredensborg": ["fredensborg", "1"],
    "Frederiksberg": ["frederiksberg", "1"],
    "Frederikssund": ["frederikssund", "1"],
    "Gentofte": ["gentofte", "1"],
    "Gladsaxe": ["gladsaxe", "1"],
    "Glostrup": ["glostrup", "1"],
    "Greve": ["greve", "1"],
    "Gribskov": ["gribskov", "1"],
    "Halsnæs": ["halsnaes", "1"],
    "Hedensted": ["hedensted", "1"],
    "Helsingør": ["helsingor", "1"],
    "Herlev": ["herlev", "1"],
    "Hillerød": ["hillerod", "1"],
    "Hjørring": ["hjoerring", "1"],
    "Horsens": ["horsens", "1"],
    "Hvidovre": ["hvidovre", "1"],
    "Høje-Taastrup": ["htk", "1"],
    "Hørsholm": ["hoersholm", "1"],
    "Jammerbugt": ["jammerbugt", "1"],
    "Kerteminde": ["kerteminde", "1"],
    "København": ["kk", "4"],
    "Køge": ["koege", "1"],
    "Lejre": ["lejre", "1"],
    "Lyngby-Taarbæk": ["ltf", "1"],
    "Mariagerfjord": ["mariagerfjord", "1"],
    "Næstved": ["naestved", "1"],
    "Odense": ["odense", "2"],
    "Odsherred": ["odsherred", "1"],
    "Randers": ["randers", "1"],
    "Ringkøbing-Skjern": ["rksk", "1"],
    "Ringsted": ["ringsted", "1"],
    "Roskilde": ["roskilde", "1"],
    "Rudersdal": ["rudersdal", "1"],
    "Rødovre": ["rk", "1"],
    "Samsø": ["samsoe", "1"],
    "Slagelse": ["slagelse", "1"],
    "Solrød": ["solrod", "1"],
    "Stevns": ["stevns", "1"],
    "Svendborg": ["svendborg", "1"],
    "Sønderborg": ["sonderborg", "1"],
    "Tårnby": ["taarnby", "1"],
    "Varde": ["varde", "1"],
    "Vejen": ["vejen", "1"],
    "Vordingborg": ["vordingborg", "1"],
}

MUNICIPALITIES_ARRAY = list(MUNICIPALITIES_LIST.keys())

ODD_EVEN_ARRAY = ["lige", "ulige"]

WEEKDAYS = ["Mandag", "Tirsdag", "Onsdag",
            "Torsdag", "Fredag", "Lørdag", "Søndag"]
WEEKDAYS_SHORT = ["Man", "Tir", "Ons", "Tor", "Fre", "Lør", "Søn"]



"""This module contains the code to get garbage data from AffaldDK."""

from __future__ import annotations

import datetime as dt
from ical.calendar_stream import IcsCalendarStream
from ical.exceptions import CalendarParseError
import json
import logging
import re
from urllib.parse import urlparse, parse_qsl

from typing import Any

import aiohttp

from .const import (
    ICON_LIST,
    MATERIAL_LIST,
    MUNICIPALITIES_LIST,
    NAME_LIST,
    NON_MATERIAL_LIST,
    NON_SUPPORTED_ITEMS,
    ODD_EVEN_ARRAY,
    SUPPORTED_ITEMS,
    WEEKDAYS,
)
from .data import PickupEvents, PickupType, AffaldDKAddressInfo

_LOGGER = logging.getLogger(__name__)


class AffaldDKNotSupportedError(Exception):
    """Raised when the municipality is not supported."""


class AffaldDKNotValidAddressError(Exception):
    """Raised when the address is not found."""


class AffaldDKNoConnection(Exception):
    """Raised when no data is received."""


class AffaldDKGarbageTypeNotFound(Exception):
    """Raised when new garbage type is detected."""


class AffaldDKAPIBase:
    """Base class for the API."""
    def __init__(self, session=None) -> None:
        """Initialize the class."""
        self.session = session

    async def async_api_request(self, url: str, body=None, as_json=True) -> dict[str, Any]:
        """Make an API request."""

        is_new_session = False
        if self.session is None:
            self.session = aiohttp.ClientSession()
            is_new_session = True

        method = 'GET'
        headers = None
        data = None
        if body:
            method = 'POST'
            headers = {"Content-Type": "application/json"}

        async with self.session.request(method, url, headers=headers, json=body) as response:
            if response.status != 200:
                if is_new_session:
                    await self.session.close()

                if response.status == 400:
                    raise AffaldDKNotSupportedError(
                        "Municipality not supported")

                if response.status == 404:
                    raise AffaldDKNotSupportedError(
                        "Municipality not supported")

                if response.status == 503:
                    raise AffaldDKNoConnection(
                        "System API is currently not available")

                raise AffaldDKNoConnection(
                    f"Error {response.status} from {url}")

            if as_json:
                data = await response.json()
            else:
                data = await response.text()
            if is_new_session:
                await self.session.close()

            return data


class NemAffaldAPI(AffaldDKAPIBase):
    # NemAffaldService API
    def __init__(self, domain, session=None):
        super().__init__(session)
        self._token = None
        self._id = None
        self.street = None
        self.base_url = f'https://nemaffaldsservice.{domain}.dk'

    @property
    async def token(self):
        if self._token is None:
            await self._get_token()
        return self._token

    async def _get_token(self):
        data = ''
        async with self.session.get(self.base_url) as response:
            data = await response.text()

        if data:
            match = re.search(r'name="__RequestVerificationToken"\s+[^>]*value="([^"]+)"', data)
            if match:
                self._token = match.group(1)

    async def get_address_id(self, zipcode, street, house_number):
        if self._id is None:
            data = {
                '__RequestVerificationToken': await self.token, 
                'SearchTerm': f"{street} {house_number}"
            }
            async with self.session.post(f"{self.base_url}/WasteHome/SearchCustomerRelation", data=data) as response:
                if len(response.history) > 1:
                    o = urlparse(response.history[1].headers['Location'])
                    self._id = dict(parse_qsl(o.query))['customerId']
        return self._id

    async def async_get_ical_data(self, customerid):
        ics_data = ''
        data = {'customerId': customerid, 'type': 'ics'}
        async with self.session.get(f"{self.base_url}/Calendar/GetICaldendar", data=data) as response:
            ics_data = await response.text()
        return ics_data


class AarhusAffaldAPI(AffaldDKAPIBase):
    # Aarhus Forsyning API
    def __init__(self, session=None):
        super().__init__(session)
        self.url_data = "https://portal-api.kredslob.dk/api/calendar/address/"
        self.url_search = "https://api.dataforsyningen.dk/adresser?kommunekode=751&q="

    async def get_address_id(self, zipcode, street, house_number):
        url = f"{self.url_search}{street.capitalize()}*"
        _LOGGER.debug("URL: %s", url)
        data: dict[str, Any] = await self.async_api_request(url)
        _result_count = len(data)
        if _result_count > 1:
            for row in data:
                if (
                    zipcode in row["adgangsadresse"]["postnummer"]["nr"]
                    and house_number == row["adgangsadresse"]["husnr"]
                ):
                    return row["kvhx"]
        return None


class OdenseAffaldAPI(AffaldDKAPIBase):
    # Odense Renovation API
    def __init__(self, session=None):
        super().__init__(session)
        self.url_data = "https://mit.odenserenovation.dk/api/Calendar/GetICalCalendar?addressNo="
        self.url_search = "https://mit.odenserenovation.dk/api/Calendar/CommunicationHouseNumbers?addressString="

    async def async_get_ical_data(self, address_id) -> dict[str, Any]:
        """Get data from iCal API."""
        url = f"{self.url_data}{address_id}"
        data = await self.async_api_request(url, as_json=False)
        data = data.replace("END:VTIMEZONE", """BEGIN:STANDARD
DTSTART:19701025T030000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=10
TZOFFSETFROM:+0200
TZOFFSETTO:+0100
END:STANDARD
BEGIN:DAYLIGHT
DTSTART:19700329T020000
RRULE:FREQ=YEARLY;BYDAY=-1SU;BYMONTH=3
TZOFFSETFROM:+0100
TZOFFSETTO:+0200
END:DAYLIGHT
END:VTIMEZONE""")
        return data

    async def get_address_id(self, zipcode, street, house_number):
        url = f"{self.url_search}{street}"
        data = await self.async_api_request(url)
        for row in data:
            if (
                zipcode in row["PostCode"]
                and house_number == row["FullHouseNumber"]
            ):
                return row["AddressNo"]
        return None


class AffaldDKAPI(AffaldDKAPIBase):
    # Renoweb API
    """Class to get data from AffaldDK."""

    def __init__(self, session=None):
        super().__init__(session)
        self.url_data = ".renoweb.dk/Legacy/JService.asmx/GetAffaldsplanMateriel_mitAffald"
        self.url_search = ".renoweb.dk/Legacy/JService.asmx/Adresse_SearchByString"

    async def get_address_id(self, municipality_url, zipcode, street, house_number):
        url = f"https://{municipality_url}{self.url_search}"
        body = {
            "searchterm": f"{street} {house_number}",
            "addresswithmateriel": 7,
        }
        # _LOGGER.debug("Municipality URL: %s %s", url, body)
        data = await self.async_api_request(url, body)
        result = json.loads(data["d"])
        # _LOGGER.debug("Address Data: %s", result)
        if "list" not in result:
            raise AffaldDKNoConnection(
                f'''AffaldDK API: {
                    result['status']['status']} - {result['status']['msg']}'''
            )

        _result_count = len(result["list"])
        _item: int = 0
        _row_index: int = 0
        if _result_count > 1:
            for row in result["list"]:
                if zipcode in row["label"] and house_number in row["label"]:
                    _item = _row_index
                    break
                _row_index += 1
        address_id = result["list"][_item]["value"]
        if address_id == "0000":
            return None
        return address_id


class GarbageCollection:
    """Class to get garbage collection data."""

    def __init__(
        self,
        municipality: str,
        session: aiohttp.ClientSession = None,
    ) -> None:
        """Initialize the class."""
        self._municipality = municipality
        # self._tzinfo = None
        self._street = None
        self._house_number = None
        self._api_data = None
        self._data = None
        self._municipality_url = None
        self._address_id = None
        for key, value in MUNICIPALITIES_LIST.items():
            if key.lower() == self._municipality.lower():
                self._api_data = value[1]
                if self._api_data == '1':
                    self._api = AffaldDKAPI(session=session)
                    self._municipality_url = value[0]
                if self._api_data == '2':
                    self._api = OdenseAffaldAPI(session=session)
                    self._municipality_url = value[0]
                elif self._api_data == '3':
                    self._api = AarhusAffaldAPI(session=session)
                    self._municipality_url = value[0]
                elif self._api_data == '4':
                    self._api = NemAffaldAPI(value[0], session=session)
                    self._municipality_url = self._api.base_url
        if session:
            self._api.session = session

    async def async_init(self) -> None:
        """Initialize the connection."""
        if self._municipality is not None:
            if self._api_data == "2":
                url = f"{self._api.url_search}{self._street}"
                await self._api.async_api_request(url)
            elif self._api_data == "3":
                url = f"{self._api.url_search}{self._street}*"
                await self._api.async_api_request(url)
            elif self._api_data == "4":
                await self._api.token
            else:
                url = f"https://{self._municipality_url}{self._api.url_search}"
                body = {
                    "searchterm": f"{self._street} {self._house_number}",
                    "addresswithmateriel": 1,
                }
                await self._api.async_api_request(url, body)

    async def get_address_id(
        self, zipcode: str, street: str, house_number: str
    ) -> AffaldDKAddressInfo:
        """Get the address id."""

        if self._municipality_url is not None:
            if self._api_data == "2":
                self._address_id = await self._api.get_address_id(zipcode, street, house_number)
            elif self._api_data == "3":
                self._address_id = await self._api.get_address_id(zipcode, street, house_number)
            elif self._api_data == "4":
                self._address_id = await self._api.get_address_id(zipcode, street, house_number)
                # self._api.street = f"{street} {house_number}"
                # self._address_id = await self._api.customerid
            else:
                self._address_id = await self._api.get_address_id(self._municipality_url, zipcode, street, house_number)

            if self._address_id is None:
                raise AffaldDKNotValidAddressError("Address not found")

            address_data = AffaldDKAddressInfo(
                self._address_id,
                self._municipality.capitalize(),
                street.capitalize(),
                house_number,
            )
            return address_data
        else:
            raise AffaldDKNotSupportedError("Cannot find Municipality")

    async def get_pickup_data(self, address_id: str) -> PickupEvents:
        """Get the garbage collection data."""

        if self._municipality_url is not None:
            pickup_events: PickupEvents = {}
            _next_pickup = dt.datetime(2030, 12, 31, 23, 59, 0)
            _next_pickup = _next_pickup.date()
            _next_pickup_event: PickupType = None
            _next_name = []
            _next_description = []

            if self._api_data == "2":
                data = await self._api.async_get_ical_data(address_id)
                try:
                    ics = IcsCalendarStream.calendar_from_ics(data)
                    for event in ics.timeline:
                        _garbage_types = split_ical_garbage_types(
                            event.summary)
                        for garbage_type in _garbage_types:
                            _pickup_date = event.start_datetime.date()
                            if _pickup_date < dt.date.today():
                                continue

                            key = get_garbage_type_from_material(
                                garbage_type, self._municipality, self._address_id
                            )
                            _pickup_event = {
                                key: PickupType(
                                    date=_pickup_date,
                                    group=key,
                                    friendly_name=NAME_LIST.get(key),
                                    icon=ICON_LIST.get(key),
                                    entity_picture=f"{key}.svg",
                                    description=garbage_type,
                                )
                            }
                            if not key_exists_in_pickup_events(pickup_events, key):
                                pickup_events.update(_pickup_event)

                            if _pickup_date is not None:
                                if _pickup_date < dt.date.today():
                                    continue
                                if _pickup_date < _next_pickup:
                                    _next_pickup = _pickup_date
                                    _next_name = []
                                    _next_description = []
                                if _pickup_date == _next_pickup:
                                    _next_name.append(NAME_LIST.get(key))
                                    _next_description.append(garbage_type)

                    if _next_name:
                        _next_pickup_event = {
                            "next_pickup": PickupType(
                                date=_next_pickup,
                                group="genbrug",
                                friendly_name=list_to_string(_next_name),
                                icon=ICON_LIST.get("genbrug"),
                                entity_picture="genbrug.svg",
                                description=list_to_string(_next_description),
                            )
                        }
                        pickup_events.update(_next_pickup_event)
                except CalendarParseError as err:
                    _LOGGER.error("Error parsing iCal data: %s", err)

            elif self._api_data == "3":
                self._address_id = address_id
                url = f"{self._api.url_data}{self._address_id}"
                data = await self._api.async_api_request(url)
                garbage_data = data[0]["plannedLoads"]
                for row in garbage_data:
                    _pickup_date = iso_string_to_date(row["date"])
                    if _pickup_date < dt.date.today():
                        continue
                    for item in row["fractions"]:
                        key = get_garbage_type_from_material(
                            item, self._municipality, self._address_id
                        )
                        _pickup_event = {
                            key: PickupType(
                                date=_pickup_date,
                                group=key,
                                friendly_name=NAME_LIST.get(key),
                                icon=ICON_LIST.get(key),
                                entity_picture=f"{key}.svg",
                                description=item,
                            )
                        }
                        if not key_exists_in_pickup_events(pickup_events, key):
                            pickup_events.update(_pickup_event)

                        if _pickup_date is not None:
                            if _pickup_date < dt.date.today():
                                continue
                            if _pickup_date < _next_pickup:
                                _next_pickup = _pickup_date
                                _next_name = []
                                _next_description = []
                            if _pickup_date == _next_pickup:
                                _next_name.append(NAME_LIST.get(key))
                                _next_description.append(item)

                if _next_name:
                    _next_pickup_event = {
                        "next_pickup": PickupType(
                            date=_next_pickup,
                            group="genbrug",
                            friendly_name=list_to_string(_next_name),
                            icon=ICON_LIST.get("genbrug"),
                            entity_picture="genbrug.svg",
                            description=list_to_string(_next_description),
                        )
                    }
                    pickup_events.update(_next_pickup_event)

            elif self._api_data == "4":
                data = await self._api.async_get_ical_data(address_id)
                try:
                    ics = IcsCalendarStream.calendar_from_ics(data)
                    for event in ics.timeline:
                        _garbage_types = split_ical_garbage_types(
                            event.summary)
                        for garbage_type in _garbage_types:
                            _pickup_date = event.start_datetime.date()
                            if _pickup_date < dt.date.today():
                                continue

                            key = get_garbage_type(garbage_type)
                            if key == garbage_type:
                                _LOGGER.warning(f"{garbage_type} is not defined in the system. Please notify the developer.")
                                continue
                            _pickup_event = {
                                key: PickupType(
                                    date=_pickup_date,
                                    group=key,
                                    friendly_name=NAME_LIST.get(key),
                                    icon=ICON_LIST.get(key),
                                    entity_picture=f"{key}.svg",
                                    description=garbage_type,
                                )
                            }
                            if not key_exists_in_pickup_events(pickup_events, key):
                                pickup_events.update(_pickup_event)

                            if _pickup_date is not None:
                                if _pickup_date < dt.date.today():
                                    continue
                                if _pickup_date < _next_pickup:
                                    _next_pickup = _pickup_date
                                    _next_name = []
                                    _next_description = []
                                if _pickup_date == _next_pickup:
                                    _next_name.append(NAME_LIST.get(key))
                                    _next_description.append(garbage_type)
                    if _next_name:
                        _next_pickup_event = {
                            "next_pickup": PickupType(
                                date=_next_pickup,
                                group="genbrug",
                                friendly_name=list_to_string(_next_name),
                                icon=ICON_LIST.get("genbrug"),
                                entity_picture="genbrug.svg",
                                description=list_to_string(_next_description),
                            )
                        }
                        pickup_events.update(_next_pickup_event)
                except CalendarParseError as err:
                    _LOGGER.error("Error parsing iCal data: %s", err)

            else:
                self._address_id = address_id
                url = f"https://{self._municipality_url}{self._api.url_data}"
                # _LOGGER.debug("URL: %s", url)
                body = {"adrid": f"{address_id}", "common": "false"}
                # _LOGGER.debug("Body: %s", body)
                data = await self._api.async_api_request(url, body)
                garbage_data = json.loads(data["d"])["list"]
                # _LOGGER.debug("Garbage Data: %s", garbage_data)

                for row in garbage_data:
                    if row["ordningnavn"] in NON_SUPPORTED_ITEMS:
                        continue

                    _pickup_date = None
                    if row["toemningsdato"] not in NON_SUPPORTED_ITEMS:
                        _pickup_date = to_date(row["toemningsdato"])
                    elif str(row["toemningsdage"]).capitalize() in WEEKDAYS:
                        _pickup_date = get_next_weekday(row["toemningsdage"])
                        _LOGGER.debug("FOUND IN TOEMNINGSDAGE")
                    elif find_weekday_in_string(row["toemningsdage"]) != "None":
                        if row["toemningsdato"] not in NON_SUPPORTED_ITEMS:
                            _weekday = find_weekday_in_string(
                                row["toemningsdage"])
                            _pickup_date = get_next_weekday(_weekday)
                        elif find_odd_even_in_string(row["toemningsdage"]) != "None":
                            _weekday = find_weekday_in_string(
                                row["toemningsdage"])
                            _odd_even = find_odd_even_in_string(
                                row["toemningsdage"])
                            _LOGGER.debug("WEEK: %s - %s", _odd_even, _weekday)
                            _pickup_date = get_next_weekday_odd_even(
                                _weekday, _odd_even)
                        else:
                            _pickup_date = get_next_year_end()
                    else:
                        continue

                    if (
                        any(
                            group in row["ordningnavn"].lower()
                            for group in [
                                "genbrug",
                                "storskrald",
                                "papir og glas/dåser",
                                "miljøkasse/tekstiler",
                            ]
                        )
                        and self._municipality.lower() == "gladsaxe"
                    ):
                        key = get_garbage_type_from_material(
                            row["materielnavn"], self._municipality, self._address_id
                        )
                    elif (
                        any(
                            group in row["ordningnavn"].lower()
                            for group in [
                                "dagrenovation",
                            ]
                        )
                        and self._municipality.lower() == "gribskov"
                    ):
                        key = get_garbage_type_from_material(
                            row["materielnavn"], self._municipality, self._address_id
                        )

                    elif any(
                        group in row["ordningnavn"].lower()
                        for group in [
                            "genbrug",
                            "papir og glas/dåser",
                            "miljøkasse/tekstiler",
                            "standpladser",
                        ]
                    ):
                        key = get_garbage_type_from_material(
                            row["materielnavn"], self._municipality, self._address_id
                        )
                    else:
                        key = get_garbage_type(row["ordningnavn"])

                    if key == row["ordningnavn"] and key != "Bestillerordning":
                        _LOGGER.warning(
                            "Garbage type [%s] is not defined in the system. Please notify the developer. Municipality: %s, Address ID: %s",
                            key,
                            self._municipality,
                            self._address_id,
                        )
                        continue

                    _pickup_event = {
                        key: PickupType(
                            date=_pickup_date,
                            group=row["ordningnavn"],
                            friendly_name=NAME_LIST.get(key),
                            icon=ICON_LIST.get(key),
                            entity_picture=f"{key}.svg",
                            description=row["materielnavn"],
                        )
                    }
                    pickup_events.update(_pickup_event)

                    if _pickup_date is not None:
                        if _pickup_date < dt.date.today():
                            continue
                        if _pickup_date < _next_pickup:
                            _next_pickup = _pickup_date
                            _next_name = []
                            _next_description = []
                        if _pickup_date == _next_pickup:
                            _next_name.append(NAME_LIST.get(key))
                            _next_description.append(row["materielnavn"])

                if _next_name:
                    _next_pickup_event = {
                        "next_pickup": PickupType(
                            date=_next_pickup,
                            group="genbrug",
                            friendly_name=list_to_string(_next_name),
                            icon=ICON_LIST.get("genbrug"),
                            entity_picture="genbrug.svg",
                            description=list_to_string(_next_description),
                        )
                    }
                    pickup_events.update(_next_pickup_event)
            return pickup_events


def to_date(datetext: str) -> dt.date:
    """Convert a date string to a datetime object."""
    if datetext == "Ingen tømningsdato fundet!":
        return None

    index = datetext.rfind(" ")
    if index == -1:
        return None
    _date = dt.datetime.strptime(f"{datetext[index+1:]}", "%d-%m-%Y")
    return _date.date()


def iso_string_to_date(datetext: str) -> dt.date:
    """Convert a date string to a datetime object."""
    if datetext == "Ingen tømningsdato fundet!":
        return None

    return dt.datetime.fromisoformat(datetext).date()


def get_garbage_type(item: str) -> str:
    """Get the garbage type."""
    # _LOGGER.debug("Affalds type: %s", item)
    for key, value in SUPPORTED_ITEMS.items():
        if item.lower() in str(value).lower():
            for entry in value:
                if item.lower() == entry.lower():
                    return key
    return item


def get_garbage_type_from_material(
    item: str, municipality: str, address_id: str
) -> str:
    """Get the garbage type from the materialnavn."""
    # _LOGGER.debug("Material: %s", item)
    for key, value in MATERIAL_LIST.items():
        if item in NON_MATERIAL_LIST:
            continue
        if item.lower() in str(value).lower():
            for entry in value:
                if item.lower() == entry.lower():
                    return key

    _LOGGER.warning(
        "Material type [%s] is not defined in the system for Genbrug. Please notify the developer. Municipality: %s, Address ID: %s",
        item,
        municipality,
        address_id,
    )
    return "genbrug"


def get_next_weekday(weekday: str) -> dt.date:

    weekdays = WEEKDAYS
    current_weekday = dt.datetime.now().weekday()
    target_weekday = weekdays.index(weekday.capitalize())
    days_ahead = (target_weekday - current_weekday) % 7
    next_date: dt.date = dt.datetime.now() + dt.timedelta(days=days_ahead)
    return next_date.date()


def get_next_weekday_odd_even(weekday: str, odd_even: str) -> dt.date:
    """Get next date for a weekday considering odd/even weeks.

    Args:
        weekday: String with weekday name
        odd_even: String with 'ulige' or 'lige' for odd/even weeks

    Returns:
        dt.date: Next date matching weekday and odd/even week criteria
    """
    weekdays = WEEKDAYS
    current_date = dt.datetime.now()
    target_weekday = weekdays.index(weekday.capitalize())

    # Find next occurrence of weekday
    days_ahead = (target_weekday - current_date.weekday()) % 7
    next_date = current_date + dt.timedelta(days=days_ahead)
    if days_ahead == 0:  # If today is the target weekday, move to next week
        next_date += dt.timedelta(days=7)

    # Check if week number matches odd/even criteria using ISO week numbers
    week_number = next_date.isocalendar()[1]
    _LOGGER.debug("Week Number: %s", week_number)
    is_odd_week = week_number % 2 == 1
    needs_odd = odd_even.lower() == 'ulige'

    # If initial date doesn't match odd/even criteria, add a week
    if is_odd_week != needs_odd:
        next_date += dt.timedelta(days=7)

    return next_date.date()


def list_to_string(list: list[str]) -> str:
    """Convert a list to a string."""
    return " | ".join(list)


def find_weekday_in_string(text: str) -> str:
    """Loop through each word in a text string and compare with another word."""
    words = text.split()
    for w in words:
        if w.capitalize() in WEEKDAYS:
            return w.capitalize()
    return "None"


def find_odd_even_in_string(text: str) -> str:
    """Loop through each word in a text string and compare with another word."""
    words = text.split()
    for w in words:
        if w.lower() in ODD_EVEN_ARRAY:
            return w.lower()
    return "None"


def get_next_year_end() -> dt.date:
    """Return December 31 of the next year."""
    today = dt.date.today()
    next_year = today.year + 1
    return dt.date(next_year, 12, 31)


def split_ical_garbage_types(text: str) -> list[str]:
    """Split a text string at every comma and ignore everything from 'på' or if it starts with 'Tømning af'."""
    if text.startswith("Tømning af"):
        text = text[len("Tømning af "):]
    if "på" in text:
        text = text.split("på")[0]
    return [item.strip() for item in text.split(",")]


def key_exists_in_pickup_events(pickup_events: PickupEvents, key: str) -> bool:
    """Check if a key exists in PickupEvents."""
    return key in pickup_events


def value_exists_in_pickup_events(pickup_events: PickupEvents, value: Any) -> bool:
    """Check if a value exists in PickupEvents."""
    return any(event for event in pickup_events.values() if event == value)
